print("import: 'qpd'")
import qpd

print("import: 'qpd_dask'")
import qpd_dask

print("import: 'qpd_pandas'")
import qpd_pandas

print("import: 'qpd_ray'")
import qpd_ray

print("import: 'qpd_test'")
import qpd_test

print("import: 'qpd_version'")
import qpd_version




from typing import Any, Dict

from qpd._parser.sql import QPDSql
from qpd._parser.utils import VisitorContext
from qpd._parser.visitors import StatementVisitor
from qpd.qpd_engine import QPDEngine
from qpd.workflow import QPDWorkflow, QPDWorkflowContext


def run_sql(
    engine: QPDEngine, sql: str, dfs: Dict[str, Any], ignore_case: bool = False
) -> Any:
    qsql = QPDSql(sql, "singleStatement", ignore_case=ignore_case)
    ctx = QPDWorkflowContext(engine, dfs)
    wf = QPDWorkflow(ctx)
    v = StatementVisitor(
        VisitorContext(
            sql=qsql,
            workflow=wf,
            dfs=wf.dfs,
        )
    )
    wf.assemble_output(v.visit(qsql.tree))
    wf.run()
    return ctx.result



from abc import ABC, abstractmethod
from builtins import bool
from typing import Any, Dict, List, Tuple

from triad.utils.assertion import assert_or_throw
from triad.utils.pandas_like import PandasLikeUtils

from qpd.dataframe import Column, DataFrame
from qpd.specs import (
    AggFunctionSpec,
    ArgumentSpec,
    IsValueSpec,
    OrderBySpec,
    WindowFunctionSpec,
)


class QPDEngine(ABC):
    def __call__(self, func_name: str, *args: Any, **kwargs: Any) -> Any:
        return getattr(self, func_name)(*args, **kwargs)

    def rename(self, col: Column, name: str) -> Column:
        return col.rename(name)

    def extract_col(self, df: DataFrame, name: str) -> Column:
        return df[name]

    def assemble_df(self, *args: Any) -> DataFrame:
        return DataFrame(*args)

    @property
    @abstractmethod
    def pl_utils(self) -> PandasLikeUtils:
        raise NotImplementedError

    @abstractmethod
    def to_df(self, obj: Any) -> DataFrame:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    def to_col(self, value: Any, name: str = "") -> Column:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    def to_native(self, df: DataFrame) -> Any:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    def is_series(self, obj: Any) -> bool:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    def case_when(self, *cols: Column) -> Column:  # pragma: no cover
        """`cols` must be in the format of
        `when1`, `value1`, ... ,`default`, and length must be >= 3.
        The reason to design the interface in this way is to simplify the translation
        from SQL to engine APIs.

        :return: [description]
        :rtype: Column
        """
        raise NotImplementedError

    @abstractmethod
    def order_by_limit(
        self, df: DataFrame, order_by: OrderBySpec, limit: int
    ) -> DataFrame:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    def group_agg(
        self,
        df: DataFrame,
        keys: List[str],
        agg_map: Dict[str, Tuple[str, AggFunctionSpec]],
    ) -> DataFrame:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    def window(  # noqa: C901
        self,
        df: DataFrame,
        func: WindowFunctionSpec,
        args: List[ArgumentSpec],
        dest_col_name: str,
    ) -> DataFrame:  # pragma: no cover
        raise NotImplementedError

    def basic_unary_arithmetic_op(self, col: Column, op: str) -> Column:
        if op == "+":
            return col
        if op == "-":
            return Column(0 - col.native)
        raise NotImplementedError(f"{op} is not supported")  # pragma: no cover

    def binary_arithmetic_op(self, col1: Column, col2: Column, op: str) -> Column:
        if op == "+":
            return Column(col1.native + col2.native)
        if op == "-":
            return Column(col1.native - col2.native)
        if op == "*":
            return Column(col1.native * col2.native)
        if op == "/":
            return Column(col1.native / col2.native)
        raise NotImplementedError(f"{op} is not supported")  # pragma: no cover

    def comparison_op(self, col1: Column, col2: Column, op: str) -> Column:
        if op == "==":
            s: Any = col1.native == col2.native
        elif op == "!=":
            s = col1.native != col2.native
        elif op == "<":
            s = col1.native < col2.native
        elif op == "<=":
            s = col1.native <= col2.native
        elif op == ">":
            s = col1.native > col2.native
        elif op == ">=":
            s = col1.native >= col2.native
        else:  # pragma: no cover
            raise NotImplementedError(f"{op} is not supported")
        return self._set_op_result_to_none(s, col1.native, col2.native)

    def binary_logical_op(self, col1: Column, col2: Column, op: str) -> Column:
        c1 = self._safe_bool(col1)
        c2 = self._safe_bool(col2)
        if op == "and":
            s: Any = c1 * c2
            # in sql, FALSE AND anything is False
            if self.is_series(s):
                s = s.mask((c1 == 0) | (c2 == 0), 0)
            elif (c1 == 0) | (c2 == 0):
                s = 0.0
        elif op == "or":
            s = c1 + c2
            # in sql, True OR anything is True
            if self.is_series(s):
                s = s.mask((c1 > 0) | (c2 > 0), 1)
            elif (c1 > 0) | (c2 > 0):
                s = 1.0
        else:  # pragma: no cover
            raise NotImplementedError(f"{op} is not supported")
        return Column(s)

    def logical_not(self, col: Column) -> Column:
        s = self._safe_bool(col.native)
        if self.is_series(s):
            nulls = s.isnull()
            s = s == 0
            s = s.mask(nulls, None)
            return Column(s)
        return Column(1.0 - s)

    def filter_df(self, df: DataFrame, cond: Column) -> DataFrame:
        c = self._safe_bool(cond)
        if self.is_series(c):
            ndf = self.to_native(df)[c > 0]
            return self.to_df(ndf)
        elif c > 0:
            return df
        else:
            ndf = self.to_native(df).head(0)
            return self.to_df(ndf)

    def is_value(self, col: Column, is_value: IsValueSpec) -> Column:
        v = is_value.value_expr
        if v == "null":
            if is_value.positive:
                return Column(col.native.isnull())
            else:
                return Column(~col.native.isnull())
        if v == "true":
            if is_value.positive:
                return Column(col.native == True)  # noqa:
            else:
                return Column(col.native != True)  # noqa:
        if v == "false":
            if is_value.positive:
                return Column(col.native == False)  # noqa:
            else:
                return Column(col.native != False)  # noqa:
        raise NotImplementedError(v)  # pragma: no cover

    def is_in(self, col: Column, *values: Any, positive: bool) -> Column:
        cols = [x for x in values if isinstance(x, Column)]
        others = [x for x in values if not isinstance(x, Column)]
        if positive:
            o: Any = col.native.isin(others)
            for c in cols:
                o = o | col.native == c.native
        else:
            o = ~col.native.isin(others)
            for c in cols:
                o = o & col.native != c.native
        if self.is_series(o) and self.is_series(col.native):
            o = o.mask(col.native.isnull(), None)
        return Column(o)

    def is_between(
        self, col: Column, lower: Column, upper: Column, positive: bool
    ) -> Column:
        ln = lower.native
        cn = col.native
        un = upper.native
        if positive:
            s: Any = (ln <= cn) & (cn <= un)
        else:
            s = (ln > cn) | (cn > un)
        if self.is_series(col.native):
            s = s.mask(col.native.isnull(), None)
        if self.is_series(lower.native):
            s = s.mask(lower.native.isnull(), None)
        if self.is_series(upper.native):
            s = s.mask(upper.native.isnull(), None)
        return Column(s)

    def drop_duplicates(self, df: DataFrame) -> DataFrame:
        ndf = self.to_native(df)
        return self.to_df(self.pl_utils.drop_duplicates(ndf))

    def union(self, df1: DataFrame, df2: DataFrame, unique: bool) -> DataFrame:
        ndf1 = self.to_native(df1)
        ndf2 = self.to_native(df2)
        return self.to_df(self.pl_utils.union(ndf1, ndf2, unique))

    def intersect(self, df1: DataFrame, df2: DataFrame, unique: bool) -> DataFrame:
        ndf1 = self.to_native(df1)
        ndf2 = self.to_native(df2)
        return self.to_df(self.pl_utils.intersect(ndf1, ndf2, unique))

    def except_df(self, df1: DataFrame, df2: DataFrame, unique: bool) -> DataFrame:
        ndf1 = self.to_native(df1)
        ndf2 = self.to_native(df2)
        return self.to_df(self.pl_utils.except_df(ndf1, ndf2, unique))

    def join(
        self, df1: DataFrame, df2: DataFrame, join_type: str, on: List[str]
    ) -> DataFrame:
        ndf1 = self.to_native(df1)
        ndf2 = self.to_native(df2)
        joined = self.pl_utils.join(ndf1, ndf2, join_type=join_type, on=on)
        return self.to_df(joined)

    def _set_op_result_to_none(self, series: Any, s1: Any, s2: Any) -> Column:
        if not self.is_series(series):
            if s1 is None or s2 is None:
                return Column(None)
            return Column(series)
        if self.is_series(s1):
            series = series.mask(s1.isnull(), None)
        if self.is_series(s2):
            series = series.mask(s2.isnull(), None)
        return Column(series)

    def _preprocess_set_op(self, df1: DataFrame, df2: DataFrame) -> Tuple[Any, Any]:
        assert_or_throw(
            len(df1) == len(df2),
            ValueError("two dataframes have different number of columns"),
        )
        ndf1 = self.to_native(df1)
        ndf2 = self.to_native(df2)
        ndf2.columns = ndf1.columns  # this is SQL behavior
        return ndf1, ndf2

    def _safe_bool(self, col: Any) -> Any:
        if self.is_series(col):
            return col.astype("f8")
        if self.is_series(col.native):
            return col.native.astype("f8")
        elif col is None:
            return float("nan")
        else:
            return float(col.native > 0)



from typing import Any, Dict, List, Tuple
from unittest import TestCase

import numpy as np
import pandas as pd
from pandasql import sqldf
from datetime import datetime, timedelta

from qpd.dataframe import DataFrame, Column
from qpd import run_sql
from qpd.qpd_engine import QPDEngine
from qpd_test.utils import assert_df_eq


class TestsBase(TestCase):
    def make_qpd_engine(self) -> QPDEngine:  # pragma: no cover
        raise NotImplementedError

    def to_native_df(self, data: Any, columns: Any) -> Any:  # pragma: no cover
        raise NotImplementedError

    def to_pandas_df(self, df: DataFrame) -> pd.DataFrame:  # pragma: no cover
        raise NotImplementedError

    def to_pandas_series(self, col: Column) -> pd.Series:  # pragma: no cover
        raise NotImplementedError

    def to_df(self, data: Any, cols: Any) -> DataFrame:
        return self.op.to_df(self.to_native_df(data, cols))

    @classmethod
    def setUpClass(cls):
        cls._qpd_engine = cls.make_qpd_engine(cls)

    @property
    def op(self) -> QPDEngine:
        return self._qpd_engine  # type: ignore

    @classmethod
    def tearDownClass(cls):
        pass

    def eq_sqlite(self, sql: str, **dfs: Any) -> None:
        self.assert_eq(dfs, sql, None, None)

    def make_rand_df(self, size: int, **kwargs: Any) -> Tuple[Any, List[str]]:
        np.random.seed(0)
        data: Dict[str, np.ndarray] = {}
        for k, v in kwargs.items():
            if not isinstance(v, tuple):
                v = (v, 0.0)
            dt, null_ct = v[0], v[1]
            if dt is int:
                s = np.random.randint(10, size=size)
            elif dt is bool:
                s = np.where(np.random.randint(2, size=size), True, False)
            elif dt is float:
                s = np.random.rand(size)  # type: ignore
            elif dt is str:
                r = [f"ssssss{x}" for x in range(10)]
                c = np.random.randint(10, size=size)
                s = np.array([r[x] for x in c])
            elif dt is datetime:
                rt = [datetime(2020, 1, 1) + timedelta(days=x) for x in range(10)]
                c = np.random.randint(10, size=size)
                s = np.array([rt[x] for x in c])
            else:
                raise NotImplementedError
            ps = pd.Series(s)
            if null_ct > 0:
                idx = np.random.choice(size, null_ct, replace=False).tolist()
                ps[idx] = None
            data[k] = ps
        pdf = pd.DataFrame(data)
        return pdf.values.tolist(), list(pdf.columns)

    def to_dfs(self, dfs: Dict[str, Any]) -> Dict[str, DataFrame]:
        res: Dict[str, DataFrame] = {}
        for k, v in dfs.items():
            if isinstance(v, DataFrame):
                res[k] = v
            elif isinstance(v, tuple):
                res[k] = self.to_df(v[0], v[1])
            else:
                res[k] = self.to_df(v, None)
        return res

    def to_pd_dfs(self, dfs: Dict[str, Any]) -> Dict[str, pd.DataFrame]:
        res: Dict[str, DataFrame] = {}
        for k, v in dfs.items():
            if isinstance(v, tuple):
                res[k] = pd.DataFrame(v[0], columns=v[1])
            elif isinstance(v, pd.DataFrame):
                res[k] = v
            else:
                raise NotImplementedError
        return res

    def assert_eq(
        self,
        dfs: Dict[str, Any],
        sql: str,
        expected: Any = None,
        expected_cols: Any = None,
    ) -> None:
        df = self.to_pandas_df(self._run(sql, self.to_dfs(dfs)))
        if expected is None:
            expected = sqldf(sql, self.to_pd_dfs(dfs))
        print("this", df)
        print(expected)
        assert_df_eq(df, expected, expected_cols)

    def assert_column_eq(self, col: Column, expected: Any) -> None:
        res = self.to_pandas_series(col)
        assert_df_eq(pd.DataFrame({"a": res}), pd.DataFrame({"a": expected}))

    def _run(self, raw_sql: str, dfs: Dict[str, Any]) -> Any:
        return run_sql(self.op, raw_sql, dfs)
